@@ -18,15 +18,18 @@
 #
 # Red Hat Author(s): Radek Vykydal <rvykydal@redhat.com>
 #
+import gi
+gi.require_version('NetworkManager', '1.0')
 
 from gi.repository import Gio, GLib
 from gi.repository import NetworkManager
-import IPy
 import struct
 import socket
-import re
+import logging
+log = logging.getLogger("anaconda")
 
 from pyanaconda.constants import DEFAULT_DBUS_TIMEOUT
+from pyanaconda.flags import flags, can_touch_runtime_system
 
 supported_device_types = [
     NetworkManager.DeviceType.ETHERNET,
@@ -35,6 +38,7 @@ supported_device_types = [
     NetworkManager.DeviceType.BOND,
     NetworkManager.DeviceType.VLAN,
     NetworkManager.DeviceType.BRIDGE,
+    NetworkManager.DeviceType.TEAM,
 ]
 
 class UnknownDeviceError(ValueError):
@@ -47,17 +51,17 @@ class UnmanagedDeviceError(Exception):
     def __str__(self):
         return self.__repr__()
 
+class DeviceNotActiveError(Exception):
+    """Device of specified name is not active"""
+    def __str__(self):
+        return self.__repr__()
+
 class PropertyNotFoundError(ValueError):
     """Property of NM object was not found"""
     def __str__(self):
         return self.__repr__()
 
-class SettingNotFoundError(ValueError):
-    """Setting of NMRemoteConnection was not found"""
-    def __str__(self):
-        return self.__repr__()
-
-class DeviceSettingsNotFoundError(ValueError):
+class SettingsNotFoundError(ValueError):
     """Settings NMRemoteConnection object was not found"""
     def __str__(self):
         return self.__repr__()
@@ -67,32 +71,53 @@ class UnknownMethodGetError(Exception):
     def __str__(self):
         return self.__repr__()
 
+# bug #1062417 e.g. for ethernet device without link
+class UnknownConnectionError(Exception):
+    """Connection is not available for the device"""
+    def __str__(self):
+        return self.__repr__()
+
+class AddConnectionError(Exception):
+    """Connection is not available for the device"""
+    def __str__(self):
+        return self.__repr__()
+
+# bug #1039006
+class BondOptionsError(AddConnectionError):
+    pass
+
 def _get_proxy(bus_type=Gio.BusType.SYSTEM,
-               flags=Gio.DBusProxyFlags.NONE,
+               proxy_flags=Gio.DBusProxyFlags.NONE,
                info=None,
                name="org.freedesktop.NetworkManager",
                object_path="/org/freedesktop/NetworkManager",
                interface_name="org.freedesktop.NetworkManager",
                cancellable=None):
-    proxy = Gio.DBusProxy.new_for_bus_sync(bus_type,
-                                           flags,
-                                           info,
-                                           name,
-                                           object_path,
-                                           interface_name,
-                                           cancellable)
+    try:
+        proxy = Gio.DBusProxy.new_for_bus_sync(bus_type,
+                                               proxy_flags,
+                                               info,
+                                               name,
+                                               object_path,
+                                               interface_name,
+                                               cancellable)
+    except GLib.GError as e:
+        if can_touch_runtime_system("raise GLib.GError", touch_live=True):
+            raise
+
+        log.error("_get_proxy failed: %s", e)
+        proxy = None
+
     return proxy
 
 def _get_property(object_path, prop, interface_name_suffix=""):
     interface_name = "org.freedesktop.NetworkManager" + interface_name_suffix
     proxy = _get_proxy(object_path=object_path, interface_name="org.freedesktop.DBus.Properties")
-    args = GLib.Variant('(ss)', (interface_name, prop))
+    if not proxy:
+        return None
+
     try:
-        prop = proxy.call_sync("Get",
-                                args,
-                                Gio.DBusCallFlags.NONE,
-                                DEFAULT_DBUS_TIMEOUT,
-                                None)
+        prop = proxy.Get('(ss)', interface_name, prop)
     except GLib.GError as e:
         if "org.freedesktop.DBus.Error.AccessDenied" in e.message:
             return None
@@ -101,7 +126,7 @@ def _get_property(object_path, prop, interface_name_suffix=""):
         else:
             raise
 
-    return prop.unpack()[0]
+    return prop
 
 def nm_state():
     """Return state of NetworkManager
@@ -109,7 +134,13 @@ def nm_state():
     :return: state of NetworkManager
     :rtype: integer
     """
-    return _get_property("/org/freedesktop/NetworkManager", "State")
+    prop = _get_property("/org/freedesktop/NetworkManager", "State")
+
+    # If this is an image/dir install assume the network is up
+    if not prop and (flags.imageInstall or flags.dirInstall):
+        return NetworkManager.State.CONNECTED_GLOBAL
+    else:
+        return prop
 
 # FIXME - use just GLOBAL? There is some connectivity checking
 # for GLOBAL in NM (nm_connectivity_get_connected), not sure if
@@ -143,14 +174,10 @@ def nm_devices():
     interfaces = []
 
     proxy = _get_proxy()
-    args = None
-    devices = proxy.call_sync("GetDevices",
-                              args,
-                              Gio.DBusCallFlags.NONE,
-                              DEFAULT_DBUS_TIMEOUT,
-                              None)
+    if not proxy:
+        return []
 
-    devices = devices.unpack()[0]
+    devices = proxy.GetDevices()
     for device in devices:
         device_type = _get_property(device, "DeviceType", ".Device")
         if device_type not in supported_device_types:
@@ -170,13 +197,21 @@ def nm_activated_devices():
     interfaces = []
 
     active_connections = _get_property("/org/freedesktop/NetworkManager", "ActiveConnections")
+    if not active_connections:
+        return []
+
     for ac in active_connections:
-        state = _get_property(ac, "State", ".Connection.Active")
+        try:
+            state = _get_property(ac, "State", ".Connection.Active")
+        except UnknownMethodGetError:
+            continue
         if state != NetworkManager.ActiveConnectionState.ACTIVATED:
             continue
         devices = _get_property(ac, "Devices", ".Connection.Active")
         for device in devices:
-            iface = _get_property(device, "Interface", ".Device")
+            iface = _get_property(device, "IpInterface", ".Device")
+            if not iface:
+                iface = _get_property(device, "Interface", ".Device")
             interfaces.append(iface)
 
     return interfaces
@@ -218,20 +253,13 @@ def nm_device_property(name, prop):
     retval = None
 
     proxy = _get_proxy()
-    args = GLib.Variant('(s)', (name,))
     try:
-        device = proxy.call_sync("GetDeviceByIpIface",
-                                  args,
-                                  Gio.DBusCallFlags.NONE,
-                                  DEFAULT_DBUS_TIMEOUT,
-                                  None)
+        device = proxy.GetDeviceByIpIface('(s)', name)
     except GLib.GError as e:
         if "org.freedesktop.NetworkManager.UnknownDevice" in e.message:
             raise UnknownDeviceError(name, e)
         raise
 
-    device = device.unpack()[0]
-
     retval = _get_property(device, prop, ".Device")
     if not retval:
         # Look in device type based interface
@@ -281,6 +309,30 @@ def nm_device_type_is_bond(name):
     """
     return nm_device_type(name) == NetworkManager.DeviceType.BOND
 
+def nm_device_type_is_team(name):
+    """Is the type of device team?
+
+       :param name: name of device
+       :type name: str
+       :return: True if type of device is TEAM, False otherwise
+       :rtype: bool
+       :raise UnknownDeviceError: if device is not found
+       :raise PropertyNotFoundError: if property is not found
+    """
+    return nm_device_type(name) == NetworkManager.DeviceType.TEAM
+
+def nm_device_type_is_bridge(name):
+    """Is the type of device bridge?
+
+       :param name: name of device
+       :type name: str
+       :return: True if type of device is BRIDGE, False otherwise
+       :rtype: bool
+       :raise UnknownDeviceError: if device is not found
+       :raise PropertyNotFoundError: if property is not found
+    """
+    return nm_device_type(name) == NetworkManager.DeviceType.BRIDGE
+
 def nm_device_type_is_vlan(name):
     """Is the type of device vlan?
 
@@ -318,6 +370,18 @@ def nm_device_hwaddress(name):
     """
     return nm_device_property(name, "HwAddress")
 
+def nm_device_perm_hwaddress(name):
+    """Return active hardware address of device ('HwAddress' property)
+
+       :param name: name of device
+       :type name: str
+       :return: active hardware address of device ('HwAddress' property)
+       :rtype: str
+       :raise UnknownDeviceError: if device is not found
+       :raise PropertyNotFoundError: if 'HwAddress' property is not found
+    """
+    return nm_device_property(name, "PermHwAddress")
+
 def nm_device_active_con_uuid(name):
     """Return uuid of device's active connection
 
@@ -440,7 +504,8 @@ def nm_device_ip_config(name, version=4):
 
     addr_list = []
     for addr, prefix, gateway in addresses:
-        # TODO - look for a library function (could have used IPy but byte order!)
+        # NOTE: There is IPy for python2, ipaddress for python3 but
+        # byte order of dbus value would need to be switched
         if version == 4:
             addr_str = nm_dbus_int_to_ipv4(addr)
             gateway_str = nm_dbus_int_to_ipv4(gateway)
@@ -488,6 +553,18 @@ def nm_device_slaves(name):
 
     return slave_ifaces
 
+def nm_hwaddr_to_device_name(hwaddr):
+    """Return device name of interface with given hardware address.
+
+        :param hwaddr: hardware address
+        :type hwaddr: str
+        :return: device name of interface having hwaddr
+        :rtype: str
+    """
+    for device in nm_devices():
+        if nm_device_perm_hwaddress(device).upper() == hwaddr.upper():
+            return device
+    return None
 
 def nm_ntp_servers_from_dhcp():
     """Return NTP servers obtained by DHCP.
@@ -516,12 +593,12 @@ def nm_ntp_servers_from_dhcp():
     return ntp_servers
 
 def _device_settings(name):
-    """Return object path of device setting.
+    """Return list of object paths of device settings
 
        :param name: name of device
        :type name: str
-       :return: path of settings, None if not found
-       :rtype: str or None
+       :return: list of paths of settings of device
+       :rtype: []
        :raise UnknownDeviceError: if device is not found
     """
     devtype = nm_device_type(name)
@@ -529,40 +606,46 @@ def _device_settings(name):
         settings = _find_settings(name, 'bond', 'interface-name')
     elif devtype == NetworkManager.DeviceType.VLAN:
         settings = _find_settings(name, 'vlan', 'interface-name')
+        if not settings:
+            # connections generated by NM from iBFT
+            _parent, _sep, vlanid = name.partition(".")
+            settings = _find_settings(int(vlanid), 'vlan', 'id')
     else:
-        try:
-            hwaddr_str = nm_device_hwaddress(name)
-        except PropertyNotFoundError:
-            settings = None
-        else:
-            settings = _settings_for_hwaddr(hwaddr_str)
+        settings = _find_settings(name, 'connection', 'interface-name')
+        if not settings:
+            try:
+                hwaddr_str = nm_device_perm_hwaddress(name)
+            except PropertyNotFoundError:
+                settings = []
+            else:
+                settings = _settings_for_hwaddr(hwaddr_str)
 
     return settings
 
 def _settings_for_ap(ssid):
-    """Return object path of wireless access point settings.
+    """Return list of object paths of wireless access point settings.
 
        :param ssid: ssid of access point
        :type ssid: str
-       :return: path of settings, None if not found
-       :rtype: str or None
+       :return: list of paths of settings of access point
+       :rtype: list
 `   """
     return _find_settings(ssid, '802-11-wireless', 'ssid',
             format_value=lambda ba: "".join(chr(b) for b in ba))
 
 def _settings_for_hwaddr(hwaddr):
-    """Return object path of settings of device specified by hw address.
+    """Return list of object paths of settings of device specified by hw address.
 
        :param hwaddr: hardware address (uppercase)
        :type hwaddr: str
-       :return: path of settings, None if not found
-       :rtype: str or None
+       :return: list of paths of settings found for hw address
+       :rtype: list
     """
     return _find_settings(hwaddr, '802-3-ethernet', 'mac-address',
             format_value=lambda ba: ":".join("%02X" % b for b in ba))
 
 def _find_settings(value, key1, key2, format_value=lambda x:x):
-    """Return object path of settings having given value of key1, key2 setting
+    """Return list of object paths of settings having given value of key1, key2 setting
 
        :param value: required value of setting
        :type value: corresponds to dbus type of setting
@@ -573,36 +656,51 @@ def _find_settings(value, key1, key2, format_value=lambda x:x):
        :param format_value: function to be called on setting value before
                             comparing
        :type format_value: function taking one argument (setting value)
-       :return: path of settings, None if not found
-       :rtype: str or None
+       :return: list of paths of settings
+       :rtype: list
     """
-    retval = None
+    retval = []
 
     proxy = _get_proxy(object_path="/org/freedesktop/NetworkManager/Settings", interface_name="org.freedesktop.NetworkManager.Settings")
 
-    args = None
-    connections = proxy.call_sync("ListConnections",
-                                  args,
-                                  Gio.DBusCallFlags.NONE,
-                                  DEFAULT_DBUS_TIMEOUT,
-                                  None)
-
-    for con in connections.unpack()[0]:
+    connections = proxy.ListConnections()
+    for con in connections:
         proxy = _get_proxy(object_path=con, interface_name="org.freedesktop.NetworkManager.Settings.Connection")
-        args = None
-        settings = proxy.call_sync("GetSettings",
-                                   args,
-                                   Gio.DBusCallFlags.NONE,
-                                   DEFAULT_DBUS_TIMEOUT,
-                                   None)
-        settings = settings.unpack()[0]
+        settings = proxy.GetSettings()
         try:
             v = settings[key1][key2]
         except KeyError:
             continue
         if format_value(v) == value:
-            retval = con
-            break
+            retval.append(con)
+
+    return retval
+
+def nm_get_settings(value, key1, key2, format_value=lambda x:x):
+    """Return settings having given value of key1, key2 setting
+
+       Returns list of settings(dicts) , None if settings were not found.
+    """
+    retval = []
+    settings_paths = _find_settings(value, key1, key2, format_value)
+    for settings_path in settings_paths:
+        proxy = _get_proxy(object_path=settings_path, interface_name="org.freedesktop.NetworkManager.Settings.Connection")
+        settings = proxy.GetSettings()
+        retval.append(settings)
+
+    return retval
+
+def nm_get_all_settings():
+    """Return all settings for logging."""
+    retval = []
+
+    proxy = _get_proxy(object_path="/org/freedesktop/NetworkManager/Settings", interface_name="org.freedesktop.NetworkManager.Settings")
+
+    connections = proxy.ListConnections()
+    for con in connections:
+        proxy = _get_proxy(object_path=con, interface_name="org.freedesktop.NetworkManager.Settings.Connection")
+        settings = proxy.GetSettings()
+        retval.append(settings)
 
     return retval
 
@@ -620,68 +718,161 @@ def nm_device_setting_value(name, key1, key2):
                 by NM
        :rtype: unpacked GDBus variant or None
        :raise UnknownDeviceError: if device is not found
-       :raise DeviceSettingsNotFoundError: if settings were not found
+       :raise SettingsNotFoundError: if settings were not found
                                            (eg for "wlan0")
     """
-    settings_path = _device_settings(name)
-    if not settings_path:
-        raise DeviceSettingsNotFoundError(name)
+    settings_paths = _device_settings(name)
+    if not settings_paths:
+        raise SettingsNotFoundError(name)
+    else:
+        settings_path = settings_paths[0]
     proxy = _get_proxy(object_path=settings_path, interface_name="org.freedesktop.NetworkManager.Settings.Connection")
-    args = None
-    settings = proxy.call_sync("GetSettings",
-                               args,
-                               Gio.DBusCallFlags.NONE,
-                               DEFAULT_DBUS_TIMEOUT,
-                               None)
-    settings = settings.unpack()[0]
+    settings = proxy.GetSettings()
     try:
         value = settings[key1][key2]
     except KeyError:
-        #raise SettingNotFoundError(key1, key2)
         value = None
     return value
 
+def nm_ap_setting_value(ssid, key1, key2):
+    """Return value of ap's setting specified by key1 and key2.
+
+       :param ssid: name of ap (ssid)
+       :type ssid: str
+       :param key1: first-level key of setting (eg "connection")
+       :type key1: str
+       :param key2: second-level key of setting (eg "uuid")
+       :type key2: str
+       :return: value of setting or None if the setting was not found
+                which means it does not exist or default value is used
+                by NM
+       :rtype: unpacked GDBus variant or None
+       :raise SettingsNotFoundError: if settings were not found
+                                           (eg for "wlan0")
+    """
+    settings_paths = _settings_for_ap(ssid)
+    if not settings_paths:
+        raise SettingsNotFoundError(ssid)
+    else:
+        settings_path = settings_paths[0]
+    proxy = _get_proxy(object_path=settings_path, interface_name="org.freedesktop.NetworkManager.Settings.Connection")
+    settings = proxy.GetSettings()
+    try:
+        value = settings[key1][key2]
+    except KeyError:
+        value = None
+    return value
+
+def nm_disconnect_device(name):
+    """Disconnect the device.
+
+       :raise UnknownDeviceError: if device is not found
+    """
+    proxy = _get_proxy()
+    try:
+        device = proxy.GetDeviceByIpIface('(s)', name)
+    except GLib.GError as e:
+        if "org.freedesktop.NetworkManager.UnknownDevice" in e.message:
+            raise UnknownDeviceError(name, e)
+        raise
+
+    device_proxy = _get_proxy(object_path=device, interface_name="org.freedesktop.NetworkManager.Device")
+    try:
+        device_proxy.Disconnect()
+    except GLib.GError as e:
+        if "org.freedesktop.NetworkManager.Device.NotActive" in e.message:
+            raise DeviceNotActiveError(name, e)
+        raise
+
 def nm_activate_device_connection(dev_name, con_uuid):
     """Activate device with specified connection.
 
-       :param dev_name: name of device
-       :type dev_name: str
+       :param dev_name: name of device or None for virtual devices
+       :type dev_name: str or None
        :param con_uuid: uuid of connection to be activated on device
        :type con_uuid: str
        :raise UnknownDeviceError: if device is not found
        :raise UnmanagedDeviceError: if device is not managed by NM
                                     or unavailable
+       :raise SettingsNotFoundError: if conneciton with given uuid was not found
+       :raise UnknownConnectionError: if connection is not available for the device
     """
 
-    proxy = _get_proxy()
-    args = GLib.Variant('(s)', (dev_name,))
+    if dev_name is None:
+        # virtual devices (eg bond, vlan)
+        device_path = "/"
+    else:
+        proxy = _get_proxy()
+        try:
+            device_path = proxy.GetDeviceByIpIface('(s)', dev_name)
+        except Exception as e:
+            if "org.freedesktop.NetworkManager.UnknownDevice" in e.message:
+                raise UnknownDeviceError(dev_name, e)
+            raise
+
+    con_paths = _find_settings(con_uuid, 'connection', 'uuid')
+    if not con_paths:
+        raise SettingsNotFoundError(con_uuid)
+
+    nm_proxy = _get_proxy()
     try:
-        device = proxy.call_sync("GetDeviceByIpIface",
-                                  args,
-                                  Gio.DBusCallFlags.NONE,
-                                  DEFAULT_DBUS_TIMEOUT,
-                                  None)
+        nm_proxy.ActivateConnection('(ooo)', con_paths[0], device_path, "/")
     except GLib.GError as e:
+        if "org.freedesktop.NetworkManager.UnmanagedDevice" in e.message:
+            raise UnmanagedDeviceError(dev_name, e)
+        elif "org.freedesktop.NetworkManager.UnknownConnection" in e.message:
+            raise UnknownConnectionError(dev_name, e)
         if "org.freedesktop.NetworkManager.UnknownDevice" in e.message:
             raise UnknownDeviceError(dev_name, e)
         raise
 
-    device_path = device.unpack()[0]
+def nm_add_connection(values):
+    """Add new connection specified by values.
 
-    con_path = _find_settings(con_uuid, 'connection', 'uuid')
+       :param values: list of settings with new values and its types
+                      [[key1, key2, value, type_str], ...]
+                      key1: first-level key of setting (eg "connection")
+                      key2: second-level key of setting (eg "uuid")
+                      value: new value
+                      type_str: dbus type of new value (eg "ay")
+       :type values: [[key1, key2, value, type_str], ...]
+                     key1: str
+                     key2: str
+                     value: object
+                     type_str: str
+    """
 
-    args = GLib.Variant('(ooo)', (con_path, device_path, "/"))
-    nm_proxy = _get_proxy()
+    settings = {}
+    for key1, key2, value, type_str in values:
+        gvalue = GLib.Variant(type_str, value)
+        if key1 not in settings:
+            settings[key1] = {}
+        settings[key1][key2] = gvalue
+
+    proxy = _get_proxy(object_path="/org/freedesktop/NetworkManager/Settings",
+                       interface_name="org.freedesktop.NetworkManager.Settings")
     try:
-        nm_proxy.call_sync("ActivateConnection",
-                            args,
-                            Gio.DBusCallFlags.NONE,
-                            DEFAULT_DBUS_TIMEOUT,
-                            None)
+        connection = proxy.AddConnection('(a{sa{sv}})', settings)
     except GLib.GError as e:
-        if "org.freedesktop.NetworkManager.UnmanagedDevice" in e.message:
-            raise UnmanagedDeviceError(dev_name, e)
+        if "bond.options: invalid option" in e.message:
+            raise BondOptionsError(e)
         raise
+    return connection
+
+def nm_delete_connection(uuid):
+    """Delete connection specified by uuid.
+
+       :param uuid: uuid of connection to be deleted
+       :type uuid: str
+       :return: True if connection was deleted, False if it was not found
+       :rtype: bool
+    """
+
+    settings_paths = _find_settings(uuid, "connection", "uuid")
+    if not settings_paths:
+        return False
+    proxy = _get_proxy(object_path=settings_paths[0], interface_name="org.freedesktop.NetworkManager.Settings.Connection")
+    proxy.Delete()
 
 def nm_update_settings_of_device(name, new_values):
     """Update setting of device.
@@ -707,12 +898,14 @@ def nm_update_settings_of_device(name, new_values):
                          value:
                          default_type_str: str
        :raise UnknownDeviceError: if device is not found
-       :raise DeviceSettingsNotFoundError: if settings were not found
+       :raise SettingsNotFoundError: if settings were not found
                                            (eg for "wlan0")
     """
-    settings_path = _device_settings(name)
-    if not settings_path:
-        raise DeviceSettingsNotFoundError(name)
+    settings_paths = _device_settings(name)
+    if not settings_paths:
+        raise SettingsNotFoundError(name)
+    else:
+        settings_path = settings_paths[0]
     return _update_settings(settings_path, new_values)
 
 def _update_settings(settings_path, new_values):
@@ -813,7 +1006,7 @@ def nm_ipv6_to_dbus_ay(address):
     :return: address in format 'ay' for NM dbus setting
     :rtype: list of bytes
     """
-    return [int(byte, 16) for byte in re.findall('.{1,2}', IPy.IP(address).strFullsize().replace(':', ''))]
+    return [int(byte) for byte in bytearray(socket.inet_pton(socket.AF_INET6, address))]
 
 def nm_ipv4_to_dbus_int(address):
     """Convert ipv4 address from string to int for dbus (switched endianess).
@@ -846,11 +1039,11 @@ def nm_dbus_int_to_ipv4(address):
     return socket.inet_ntop(socket.AF_INET, struct.pack('=L', address))
 
 def test():
-    print "NM state: %s:" % nm_state()
-    print "NM is connected: %s" % nm_is_connected()
+    print("NM state: %s:" % nm_state())
+    print("NM is connected: %s" % nm_is_connected())
 
-    print "Devices: %s" % nm_devices()
-    print "Activated devices: %s" % nm_activated_devices()
+    print("Devices: %s" % nm_devices())
+    print("Activated devices: %s" % nm_activated_devices())
 
     wireless_device = ""
 
@@ -858,108 +1051,108 @@ def test():
     devs.append("nonexisting")
     for devname in devs:
 
-        print devname
+        print(devname)
 
         try:
             devtype = nm_device_type(devname)
         except UnknownDeviceError as e:
-            print "     %s" % e
+            print("     %s" % e)
             devtype = None
         if devtype == NetworkManager.DeviceType.ETHERNET:
-            print "     type %s" % "ETHERNET"
+            print("     type %s" % "ETHERNET")
         elif devtype == NetworkManager.DeviceType.WIFI:
-            print "     type %s" % "WIFI"
+            print("     type %s" % "WIFI")
             wireless_device = devname
 
         try:
-            print "     Wifi device: %s" % nm_device_type_is_wifi(devname)
+            print("     Wifi device: %s" % nm_device_type_is_wifi(devname))
         except UnknownDeviceError as e:
-            print "     %s" % e
+            print("     %s" % e)
 
         try:
             hwaddr = nm_device_hwaddress(devname)
-            print "     HwAaddress: %s" % hwaddr
+            print("     HwAaddress: %s" % hwaddr)
         except ValueError as e:
-            print "     %s" % e
+            print("     %s" % e)
             hwaddr = ""
 
         try:
-            print "     Carrier: %s" % nm_device_carrier(devname)
+            print("     Carrier: %s" % nm_device_carrier(devname))
         except ValueError as e:
-            print "     %s" % e
+            print("     %s" % e)
 
         try:
-            print "     IP4 config: %s" % nm_device_ip_config(devname)
-            print "     IP6 config: %s" % nm_device_ip_config(devname, version=6)
-            print "     IP4 addrs: %s" % nm_device_ip_addresses(devname)
-            print "     IP6 addrs: %s" % nm_device_ip_addresses(devname, version=6)
-            print "     Udi: %s" % nm_device_property(devname, "Udi")
+            print("     IP4 config: %s" % nm_device_ip_config(devname))
+            print("     IP6 config: %s" % nm_device_ip_config(devname, version=6))
+            print("     IP4 addrs: %s" % nm_device_ip_addresses(devname))
+            print("     IP6 addrs: %s" % nm_device_ip_addresses(devname, version=6))
+            print("     Udi: %s" % nm_device_property(devname, "Udi"))
         except UnknownDeviceError as e:
-            print "     %s" % e
+            print("     %s" % e)
 
         if devname in nm_devices():
             try:
-                print "     Nonexisting: %s" % nm_device_property(devname, "Nonexisting")
+                print("     Nonexisting: %s" % nm_device_property(devname, "Nonexisting"))
             except PropertyNotFoundError as e:
-                print "     %s" % e
+                print("     %s" % e)
         try:
-            print "     Nonexisting: %s" % nm_device_property(devname, "Nonexisting")
+            print("     Nonexisting: %s" % nm_device_property(devname, "Nonexisting"))
         except ValueError as e:
-            print "     %s" % e
+            print("     %s" % e)
 
         try:
-            print "     Settings: %s" % _device_settings(devname)
+            print("     Settings: %s" % _device_settings(devname))
         except UnknownDeviceError as e:
-            print "     %s" % e
+            print("     %s" % e)
         try:
-            print "     Settings for hwaddr %s: %s" % (hwaddr, _settings_for_hwaddr(hwaddr))
+            print("     Settings for hwaddr %s: %s" % (hwaddr, _settings_for_hwaddr(hwaddr)))
         except UnknownDeviceError as e:
-            print "     %s" % e
+            print("     %s" % e)
         try:
-            print "     Setting value %s %s: %s" % ("ipv6", "method", nm_device_setting_value(devname, "ipv6", "method"))
+            print("     Setting value %s %s: %s" % ("ipv6", "method", nm_device_setting_value(devname, "ipv6", "method")))
         except ValueError as e:
-            print "     %s" % e
+            print("     %s" % e)
         try:
-            print "     Setting value %s %s: %s" % ("ipv7", "method", nm_device_setting_value(devname, "ipv7", "method"))
+            print("     Setting value %s %s: %s" % ("ipv7", "method", nm_device_setting_value(devname, "ipv7", "method")))
         except ValueError as e:
-            print "     %s" % e
+            print("     %s" % e)
 
     ssid = "Red Hat Guest"
-    print "Settings for AP %s: %s" % (ssid, _settings_for_ap(ssid))
+    print("Settings for AP %s: %s" % (ssid, _settings_for_ap(ssid)))
     ssid = "nonexisting"
-    print "Settings for AP %s: %s" % (ssid, _settings_for_ap(ssid))
+    print("Settings for AP %s: %s" % (ssid, _settings_for_ap(ssid)))
 
     devname = devs[0]
     key1 = "connection"
     key2 = "autoconnect"
     original_value = nm_device_setting_value(devname, key1, key2)
-    print "Value of setting %s %s: %s" % (key1, key2, original_value)
+    print("Value of setting %s %s: %s" % (key1, key2, original_value))
     # None means default in this case, which is true
     if original_value in (None, True):
         new_value = False
     else:
         new_value = True
 
-    print "Updating to %s" % new_value
+    print("Updating to %s" % new_value)
     nm_update_settings_of_device(devname, [[key1, key2, new_value, None]])
-    print "Value of setting %s %s: %s" % (key1, key2, nm_device_setting_value(devname, key1, key2))
+    print("Value of setting %s %s: %s" % (key1, key2, nm_device_setting_value(devname, key1, key2)))
     nm_update_settings_of_device(devname, [[key1, key2, original_value, None]])
-    print "Value of setting %s %s: %s" % (key1, key2, nm_device_setting_value(devname, key1, key2))
+    print("Value of setting %s %s: %s" % (key1, key2, nm_device_setting_value(devname, key1, key2)))
     nm_update_settings_of_device(devname, [[key1, key2, original_value, "b"]])
-    print "Value of setting %s %s: %s" % (key1, key2, nm_device_setting_value(devname, key1, key2))
+    print("Value of setting %s %s: %s" % (key1, key2, nm_device_setting_value(devname, key1, key2)))
 
     nm_update_settings_of_device(devname, [[key1, "nonexisting", new_value, None]])
     nm_update_settings_of_device(devname, [["nonexisting", "nonexisting", new_value, None]])
     try:
         nm_update_settings_of_device("nonexixting", [[key1, key2, new_value, None]])
     except UnknownDeviceError as e:
-        print "%s" % e
+        print("%s" % e)
 
     if wireless_device:
         try:
             nm_update_settings_of_device(wireless_device, [[key1, key2, new_value, None]])
-        except DeviceSettingsNotFoundError as e:
-            print "%s" % e
+        except SettingsNotFoundError as e:
+            print("%s" % e)
 
     #nm_update_settings_of_device(devname, [["connection", "id", "test", None]])
     #nm_update_settings_of_device(devname, [["connection", "timestamp", 11111111, None]])
